/****************************************************************************
* Matching Cards for Nintendo Wii
* Version: 0.2
*
* Author: teknecal
* Created: 07/06/2008
* Last Modified: 14/06/2008
* 
****************************************************************************/
#include <gccore.h>
#include <gcmodplay.h> 
#include <ogcsys.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <time.h>
#include <ogc/pad.h>
#include <ogc/system.h>
#include <jpeg/jpgogc.h>
#include <wiiuse/wpad.h>

// Graphics
#include "mouse.h"
extern char logo_picdata[];
extern int  logo_piclength;
extern char cards_picdata[];
extern int  cards_piclength;
extern char timeleft_picdata[];
extern int  timeleft_piclength;
extern char time_picdata[];
extern int  time_piclength;

JPEGIMG logo_image, cards_image, timeleft_image, time_image;

// Sounds
#include "card_flip_mod.h"
#include "next_level_mod.h"
#include "time_over_mod.h"

// Modplay
static MODPlay play;

// 2D Video Globals
static u32 *xfb[2] = { NULL, NULL };
static GXRModeObj *vmode;

// Wiimote IR
ir_t ir;

// Vars
bool ingame = false;
bool menu_running = true;
int level = 1;
int mode = 2;
int pairs = 0;
int pairs_to_win = 1;

int selectedx = -1;
int selectedy = -1;
int selectedcard = 0;
int selectedx2 = -1;
int selectedy2 = -1;
int selectedcard2 = 0;
int timetowin = 0;
int soundtimer = 0;
int ir_xcorrection = -150;
int ir_ycorrection = 0;

// Cards defines the cards type, 0 for no card, 1-18 for card images
int cards[] = { 
				0, 0, 0, 0, 0, 0, 0, 0, 0,
				0, 0, 0, 0, 0, 0, 0, 0, 0,
				0, 0, 0, 0, 0, 0, 0, 0, 0,
				0, 0, 0, 0, 0, 0, 0, 0, 0};

// Shows the cards if there is a 1
int cards_shown[] = {
				0, 0, 0, 0, 0, 0, 0, 0, 0,
				0, 0, 0, 0, 0, 0, 0, 0, 0,
				0, 0, 0, 0, 0, 0, 0, 0, 0,
				0, 0, 0, 0, 0, 0, 0, 0, 0};


// Initialise video and controls
void Initialise() {
	
	// Initialise the video system
	VIDEO_Init();
	
	// This function initialises the attached controllers
	WPAD_Init();
	
	// This gives us access to the accelerometers and infrared.
	WPAD_SetDataFormat(WPAD_CHAN_0, WPAD_FMT_BTNS_ACC_IR);
	
	// Initialise the audio
	AUDIO_Init(NULL);
	MODPlay_Init(&play);
	
	// Obtain the preferred video mode from the system
	// This will correspond to the settings in the Wii menu
	vmode = VIDEO_GetPreferredMode(NULL);
	
	// Allocate memory for the display in the uncached region
	xfb[0] = MEM_K0_TO_K1(SYS_AllocateFramebuffer(vmode));
	xfb[1] = MEM_K0_TO_K1(SYS_AllocateFramebuffer(vmode));
	
	// Initialise the console, required for printf
	console_init(xfb[0],20,20,vmode->fbWidth,vmode->xfbHeight,vmode->fbWidth*VI_DISPLAY_PIX_SZ);
	
	// Set up the video registers with the chosen mode
	VIDEO_Configure(vmode);
	
	// Tell the video hardware where our display memory is
	VIDEO_SetNextFramebuffer(xfb[0]);
	
	// Make the display visible
	VIDEO_SetBlack(FALSE);

	// Flush the video register changes to the hardware
	VIDEO_Flush();

	// Wait for Video setup to complete
	VIDEO_WaitVSync();
	if(vmode->viTVMode&VI_NON_INTERLACE) VIDEO_WaitVSync();
	
}

// Quick Vsync refresh function
void SetScreen() {
	VIDEO_SetNextFramebuffer(xfb[0]);
	VIDEO_Flush();
	VIDEO_WaitVSync();
}

// Clears the screen. Sets screen to black.
void clearscreen() {
	int x,y;
	for(y=64;y<486;y++) // We want to keep our logo, so don't clear all the screen
		for(x=0;x<320;x++)
			xfb[0][x+320*y]=COLOR_BLACK;
}

// Drawpic function specific for our mouse as it has transparency
void drawpic (int x,int y, int height, int width, unsigned int data[]) {
	int i,j;
	for(i=1;i<=width-1;i++)
		for(j=0;j<=height-1;j++)
			// We'll make our transparency colour blue, so any blue in the image won't show
			if ((data[i+width*j]!=0x41ea415b) && (data[i+width*j]!=0x35f13559))
				xfb[0][(i+x)+320*(j+16+y)]=data[i+width*j];
}


// Display a JPEG image
void display_jpeg(JPEGIMG jpeg, int offsetx, int offsety) {

	unsigned int *jpegout = (unsigned int *) jpeg.outbuffer;
	
	int i,j;
	for(i=0;i<=(jpeg.width-(jpeg.width/2)-2);i++)
		for(j=1;j<=jpeg.height-2;j++) 
			xfb[0][(i+offsetx)+320*(j+16+offsety)]=jpegout[i+(jpeg.width-(jpeg.width/2))*j];
}


// Display a JPEG image within the offset given and show only the card selected by picx/y
// This function is only to be used with our cards jpeg file
void display_card_jpeg(JPEGIMG jpeg, int offsetx, int offsety, int picx, int picy) {

	unsigned int *jpegout = (unsigned int *) jpeg.outbuffer;
	
	int pictx = (picx - 1) * 25 + (picx - 1);
	int pictx2 = 234 - picx * 25 - (picx - 1);
	int picty = (picy - 1) * 82;
	int picty2 = (81 * (5 - picy)) - picy + 4;
	offsetx = offsetx - pictx - 29;
	offsety = offsety - picty;
	
	int i,j;
	for(i=pictx;i<=(jpeg.width-(jpeg.width/2)-pictx2);i++)
		for(j=picty;j<=(jpeg.height-2)-picty2;j++) 
			xfb[0][(i+offsetx)+320*(j+16+offsety)]=jpegout[i+(jpeg.width-(jpeg.width/2))*j];
}


// Reset the cards choosen
void reset_temp_cards(int cards) {
	if (cards == 1 || cards == 0) {
		selectedx = -1;
		selectedy = -1;
		selectedcard = 0;
	}
	if (cards == 2 || cards == 0) {
		selectedx2 = -1;
		selectedy2 = -1;
		selectedcard2 = 0;
	}
}

// Display the cards (and logo)
void display_cards() {
	int xcounter = 1;
	int ycounter = 0;
	int a;
	
	for (a = 0; a < 36; a++) {
		if (cards[a] != 0) {
			if (cards_shown[a] != 0) {
				if (cards[a] <= 9) {
					display_card_jpeg(cards_image, (35 * xcounter), 70 + (95 * ycounter), cards[a], 1);
				}
				else if (cards[a] >= 10 && cards[a] <= 18) {
					display_card_jpeg(cards_image, (35 * xcounter), 70 + (95 * ycounter), cards[a] - 9, 2);
				}
				else if (cards[a] >= 20 && cards[a] <= 28) {
					display_card_jpeg(cards_image, (35 * xcounter), 70 + (95 * ycounter), cards[a] - 19, 3);
				}
				else if (cards[a] >= 29 && cards[a] <= 37) {
					display_card_jpeg(cards_image, (35 * xcounter), 70 + (95 * ycounter), cards[a] - 28, 4);
				}
			}
			else {
				display_card_jpeg(cards_image, (35 * xcounter), 70 + (95 * ycounter), 9, 5);
			}
		}
		
		if (xcounter == 9) {
			xcounter = 0;
			ycounter++;
		}
		xcounter++;
	}
}

// Display the timeleft (1 block each 2 seconds)
void display_timeleft() {
	display_jpeg(timeleft_image, 4, 51);
	
	int t;
	int timeleft = timetowin - time(0);
	for (t = 0; t < timeleft / 2; t++) {			
		display_jpeg(time_image, 50 + (t*2), 54);
	}
}

// Create level creates a random array with the number of pairs between the rows specififed and a time limit
void create_level(int xrow, int xrow2, int pairs, int time_win) {

	// Randomise the sequence
	srand((unsigned)time(0)); 
	
	// Set the time for the level
	timetowin = time(0) + time_win;

	// Set the number of pairs to win the game
	pairs_to_win = pairs;
	
	int q;
	int seq [(pairs*2)];
	int cardseq[(pairs*2)];
	int paircounter = 1;
	
	// Clear the cards shown
	for (q=0; q<36; q++) {
		cards_shown[q] = 0;
		clearscreen();
		display_cards();
		usleep(20000);
		SetScreen(); 
	}
	
	// We'll put the pairs into different parts of the array and make sure that we don't use the same pairs
	for(q=0; q<pairs; q++){ 
		seq[q] = seq[(q+pairs)] = paircounter;
		paircounter++;
		//printf("%i %i %i\n", q, seq[q],seq[(q+pairs)]); 
	}
	
	// For debugging	
	//for(q=0; q<(pairs*2); q++){ 
		//printf("seq = %i %i\n", q, seq[q] ); 
	//}
	
	// Put all cards in other array to 0
	for(q=0; q<(pairs*2); q++){ 
		cardseq[q] = 0;
	}
	
	// Mix up the array
	int test = 1;
	int ran = 0;
	for(q=0; q<(pairs*2); q++) { 
		while (test != 0) {
			ran = (rand() % (pairs*2));
			test = cardseq[ran];
			//printf("test = %i, ran = %i \n", test, ran);
		}
		
		cardseq[ran] = seq[q];
		test = 1;
		ran = 0;
		
		//printf("%i %i\n", q, cardseq[q]);
	}
	
	for(q=0; q<(pairs*2); q++){ 
		//printf("%i %i\n", q, cardseq[q]);
	}
	
	//printf("Sequence is %i \n",seq[1]);
	
	
	int xcounter = 0;
	int ycounter = 0;
	int a;
	int cardcounter = 0;

	// Assign our cards to the cards array (between certain rows)
	for (a = 0; a < 36; a++) {
		if (xcounter >= xrow && xcounter <= xrow2) {
			cards[a] = cardseq[cardcounter];
			//printf("cards %i %i \n",a,cardseq[cardcounter]);
			cardcounter++;
		}
		else {
			cards[a] = 0;
		}
		
		if (xcounter == 8) {
			xcounter = -1;
			ycounter++;
		}
		xcounter++;	
	}
}

// Make the next level
void next_level(int level_create) {
	pairs = 0;
	
	// Easy
	if (mode == 1) {
		switch(level_create) {
			case 1:
				create_level(3,5,6,80);
				break;
			case 2:
				create_level(2,5,8,100);
				break;
			case 3:
				create_level(2,6,10,120);
				break;
			case 4:
				create_level(1,6,12,140);
				break;
			case 5:
				create_level(1,7,14,160);
				break;
			case 6:
				create_level(0,7,16,180);
				break;
			case 7:
				create_level(0,8,18,200); 
				break;
		}
	}
	
	// Normal
	else if (mode == 2) {
		switch(level_create) {
			case 1:
				create_level(3,5,6,55);
				break;
			case 2:
				create_level(2,5,8,65);
				break;
			case 3:
				create_level(2,6,10,75);
				break;
			case 4:
				create_level(1,6,12,85);
				break;
			case 5:
				create_level(1,7,14,95);
				break;
			case 6:
				create_level(0,7,16,120);
				break;
			case 7:
				create_level(0,8,18,140); 
				break;
		}
	}
	
	// Hard
	else {
		switch(level_create) {
			case 1:
				create_level(3,5,6,25);
				break;
			case 2:
				create_level(2,5,8,30);
				break;
			case 3:
				create_level(2,6,10,35);
				break;
			case 4:
				create_level(1,6,12,40);
				break;
			case 5:
				create_level(1,7,14,45);
				break;
			case 6:
				create_level(0,7,16,55);
				break;
			case 7:
				create_level(0,8,18,60); 
				break;
		}	
	}
}

// Adds a card using x/y co-ordinate and returns the card type
int add_card_shown (int x, int y) {
	int xcounter = 0;
	int ycounter = 0;
	int a;
	int selected_card = 0;
	
	for (a = 0; a < 36; a++) {
		if (x == xcounter && y == ycounter) {
			if (cards[a] != 0 && cards_shown[a] != 1) {
				cards_shown[a] = 1;
				selected_card = cards[a];
			}
			//printf("%i %i\n",cards[a],cards_shown[a]);
		}
		
		if (xcounter == 8) {
			xcounter = -1;
			ycounter++;
		}
		xcounter++;
	}
	
	return selected_card;
}

// Removes the card shown on the x/y co-ordinate
void remove_card_shown (int x, int y) {
	int xcounter = 0;
	int ycounter = 0;
	int a;
	
	for (a = 0; a < 36; a++) {
		if (x == xcounter && y == ycounter) {
			cards_shown[a] = 0;
		}
		
		if (xcounter == 8) {
			xcounter = -1;
			ycounter++;
		}
		xcounter++;
	}
}

// Show the cards before starting each level
void show_cards() {
	int card;
	
	for (card=0; card<36; card++) {
		cards_shown[card] = 1;
		clearscreen();
		display_cards();
		usleep(30000);
		SetScreen();
	}
	
	usleep(2000000);
	
	for (card=0; card<36; card++) {
		cards_shown[card] = 0;
	}
	
	clearscreen();
	display_cards();
	SetScreen();
}

// A quicker way then calling all the below for animation
void animate_display() {
	clearscreen();
	display_cards();
	usleep(25000);
	SetScreen();
	
	WPAD_ScanPads();
	u32 pressed = WPAD_ButtonsDown(0);
	
	if (pressed & WPAD_BUTTON_A) {
		menu_running = false;		
	}

	// Return to the launcher application via exit
	if (pressed & WPAD_BUTTON_HOME) { exit(0); }
}

// Animate cards when waiting to start game
// There's probably an easier way to do this...
void animate_cards() {

	menu_running = true;
	
	int card;
	int card_type = 1;
	
	// If coming back from the game, clear all shown cards
	for (card=0; card<36; card++) {
		cards_shown[card] = 0;
	}
	
	for (card=0; card<36; card++) {
		cards[card] = card_type;
		card_type++;
		if (card_type == 18) { card_type = 1; }
	}
	
	for (card=0; card<9; card++) {
		cards_shown[card] = 1;
		animate_display();
	}
	
	// It's easier this way, for me at least...
	cards_shown[17] = 1;
	animate_display();
	
	cards_shown[26] = 1;
	animate_display();
	
	cards_shown[35] = 1;
	animate_display();
	
	for (card=35; card>26; card--) {
		cards_shown[card] = 1;
		animate_display();
	}
	
	cards_shown[18] = 1;
	animate_display();
	
	cards_shown[9] = 1;
	animate_display();
	
	// Press A to Play text
	cards[10] = 20;
	cards[11] = 21;
	cards[12] = 22;
	cards[13] = 23;
	cards[14] = 23;
	
	cards[16] = 28;
	
	cards[19] = 25;
	cards[20] = 26;
	
	cards[22] = 20;
	cards[23] = 27;
	cards[24] = 28;
	cards[25] = 29;
	
	for (card=10; card<17; card++) {
		cards_shown[card] = 1;
		animate_display();
	}
	
	for (card=25; card>18; card--) {
		cards_shown[card] = 1;
		animate_display();
	}
	
	card_type = 1;
	
	while (menu_running == true) { 
		int s;
		for (s=0; s<80; s++) {
			usleep(100000);
			
			WPAD_ScanPads();
			u32 pressed = WPAD_ButtonsDown(0);
			
			if (pressed & WPAD_BUTTON_A) {
				menu_running = false;
				break;
			}
			
			// Return to the launcher application via exit
			if (pressed & WPAD_BUTTON_HOME) { exit(0); }
		}
		
		if (menu_running == true) {
			// Change the cards every few seconds
			for (card=0; card<36; card++) {
				card_type = cards[card];
				if (card_type < 19) { 
					card_type++; 
				}
				
				if (card_type == 18) { 
					card_type = 1; 
				}
				cards[card] = card_type;
				animate_display();
			}
		}
	}
}

// Show the game over screen
void show_gameover() {
	int card;

	// You Win! text	
	int gameovercards[] = { 29, 26, 34, 35, 36, 32, 37 };
	int gameovercount = 0;
	for (card=0; card<36; card++) {
		if ((card >= 10 && card <= 12) || (card >= 22 && card <= 25)) {
			cards_shown[card] = 0;
			animate_display();
			
			cards[card] = gameovercards[gameovercount];
			cards_shown[card] = 1;
			
			gameovercount++;
		}
		animate_display();
	}
	
	usleep(2000000);
}

// Menu to select difficulty
void menu() {

	menu_running = true;
	
	int card;
	
	// Clear all shown cards
	for (card=0; card<36; card++) {
		cards[card] = 1;
		cards_shown[card] = 0;
	}
	
	// Mode, Easy, Normal, Hard text
	cards[2] = 30;
	cards[3] = 26;
	cards[4] = 31;
	cards[5] = 22;
	
	cards[12] = 22;
	cards[13] = 28;
	cards[14] = 23;
	cards[15] = 29;
	
	cards[21] = 32;
	cards[22] = 26;
	cards[23] = 21;
	cards[24] = 30;
	cards[25] = 28;
	cards[26] = 27;
	
	cards[30] = 33;
	cards[31] = 28;
	cards[32] = 21;
	cards[33] = 31;
	
	// Show the text
	for (card=0; card<36; card++) {
		if ((card >= 2 && card <= 5) || (card >= 12 && card <= 15) || (card >= 21 && card <= 26) || (card >= 30 && card <= 33)) {
			cards_shown[card] = 1;
		}
		animate_display();
	}
	
	int menuy = 2;
	
	while (menu_running == true) { 
		
		clearscreen();
		display_cards();
		
		// Show selection
		display_card_jpeg(cards_image, (35 * 3), 70 + (95 * menuy), 8, 5);
		
		WPAD_ScanPads();
		u32 pressed = WPAD_ButtonsDown(0);
		
		if ((pressed & WPAD_BUTTON_UP) && menuy > 1) {
			menuy--;
			usleep(100000);
		}
		else if ((pressed & WPAD_BUTTON_DOWN) && menuy < 3) {
			menuy++;
			usleep(100000);
		}
		
		if (pressed & WPAD_BUTTON_A) {
			mode = menuy;
			menu_running = false;
			break;
		}
		
		// Return to the launcher application via exit
		if (pressed & WPAD_BUTTON_HOME) { exit(0); }
		
		SetScreen();
	}	
}

// Reset the variables
void restart_game() {
	ingame = true;
	level = 1;
	reset_temp_cards(0);
	next_level(level);
	show_cards();
}

// Main loop
void game() {

	// Logo
	display_jpeg(logo_image, 50, 0);
	
	while (1) {
		while (ingame == true) {
			WPAD_ScanPads();
			u32 pressed = WPAD_ButtonsDown(0);
			
			// Movement
			WPAD_IR(0, &ir);
			
			// IR X & Y Correction
			ir.sx = ir.sx + ir_xcorrection;
			if (pressed & WPAD_BUTTON_LEFT) { 
				ir_xcorrection -= 50;
			}
			else if (pressed & WPAD_BUTTON_RIGHT) { 
				ir_xcorrection += 50;
			}
			
			ir.sy = ir.sy + ir_ycorrection;
			if (pressed & WPAD_BUTTON_UP) { 
				ir_ycorrection -= 25;
			}
			else if (pressed & WPAD_BUTTON_DOWN) { 
				ir_ycorrection += 25;
			}
			
			// Card selected?
			if (pressed & WPAD_BUTTON_A) { 
				
				// Set the regions (cards) to which we only respond to
				int xaxis=0;
				int yaxis=0;
				for (xaxis=0; xaxis<9; xaxis++) {
					for (yaxis=0; yaxis<4; yaxis++) {
						if ((ir.sx >= (14 + 70 * xaxis) && ir.sx <= (62 + 70 * xaxis) + 2) && (ir.sy >= (88 + 94 * yaxis) && ir.sy <= (166 + 94 * yaxis) + 2)) { 
							
							// Is this the first card selected?
							if (selectedx == -1 && selectedy == -1) {
								selectedx = ir.sx / 70; 
								selectedy = (ir.sy + 15) / 95 - 1; 
								
								
								// Add the card to the cards shown and get the card type, reset vars if card is already shown
								selectedcard = add_card_shown(selectedx,selectedy);
								if (selectedcard == 0) { 
									reset_temp_cards(1);
								}
								else {
									MODPlay_SetMOD(&play, card_flip_mod);
									MODPlay_Start(&play);
								}
							}
							else {
								selectedx2 = ir.sx / 70; 
								selectedy2 = (ir.sy + 15) / 95 - 1;
								
								selectedcard2 = add_card_shown(selectedx2,selectedy2);
								if (selectedcard2 == 0) {
									reset_temp_cards(2);
								}
								else {
									MODPlay_SetMOD(&play, card_flip_mod);
									MODPlay_Start(&play);
								}
							}
							
							//printf("x = %i, y = %i \n",x,y);
							//printf("sx = %i, sy = %i \n",selectedx,selectedy);
							//usleep(500000);
						}
					}
				}
				soundtimer = time(0) + 1;
			}
			
			// Quit to Main menu
			if (pressed & WPAD_BUTTON_PLUS) { 
				ingame = false;
			}
			
			clearscreen(); 
			display_timeleft();
			display_cards();
			
			
			// Check if selected cards match
			if (selectedx != -1 && selectedy != -1 && selectedx2 != -1 && selectedy2 != -1) {
				if ((selectedcard != 0 && selectedcard2 != 0) && selectedcard == selectedcard2) {
					pairs++;
					reset_temp_cards(0);
				}
				else {
					// Sleep a little
					usleep(900000);
					
					// Remove shown cards if they don't match
					remove_card_shown(selectedx, selectedy);
					remove_card_shown(selectedx2, selectedy2);
					reset_temp_cards(0);
				}
			}
			
			// User's cursor
			if (ir.sx > 10 && ir.sx < 630 && ir.sy > 80 && ir.sy < 460) {
				drawpic (ir.sx / 2 - 6, ir.sy - 18, 24, 12, mouse);
			}
			
			// Change to next level if user has won and show them the cards quickly
			if (pairs == pairs_to_win) {
				MODPlay_SetMOD(&play, next_level_mod);
				MODPlay_Start(&play);
				level++;
				reset_temp_cards(0);
				
				// Game over?
				if (level >= 8) {
					ingame = false;
					show_gameover();
				}
				else {
					next_level(level);
					show_cards();
				}
				
				MODPlay_Stop(&play);
			}
			
			// Time ran out? Restart the level automatically
			int timeleft = timetowin - time(0);
			if (timeleft <= 0) {
				MODPlay_SetMOD(&play, time_over_mod);
				MODPlay_Start(&play);
				reset_temp_cards(0);
				next_level(level);
				show_cards();
				MODPlay_Stop(&play);
			}
			
			// For stopping the sound after A button is pressed on a card
			if (soundtimer != 0 && soundtimer < time(0)) {
				MODPlay_Stop(&play);
				soundtimer = 0;
			}
			
			// Return to the launcher application via exit
			if (pressed & WPAD_BUTTON_HOME) { exit(0); }
			
			SetScreen();
		}
		
		// First run or restart? Then go back to menu
		animate_cards();
		menu();
		restart_game();
	}
}

int main () {

	Initialise();
	
	// Decompress the Jpegs
	memset(&logo_image, 0, sizeof(JPEGIMG));
	logo_image.inbuffer = logo_picdata;
	logo_image.inbufferlength = logo_piclength;
	JPEG_Decompress(&logo_image);
	
	memset(&cards_image, 0, sizeof(JPEGIMG));
	cards_image.inbuffer = cards_picdata;
	cards_image.inbufferlength = cards_piclength;
	JPEG_Decompress(&cards_image);
	
	memset(&timeleft_image, 0, sizeof(JPEGIMG));
	timeleft_image.inbuffer = timeleft_picdata;
	timeleft_image.inbufferlength = timeleft_piclength;
	JPEG_Decompress(&timeleft_image);
	
	memset(&time_image, 0, sizeof(JPEGIMG));
	time_image.inbuffer = time_picdata;
	time_image.inbufferlength = time_piclength;
	JPEG_Decompress(&time_image);
	
	game();
	
	return 0;
}
